import { TestBed, inject } from '@angular/core/testing';
import { HttpClientTestingModule, HttpTestingController } from '@angular/common/http/testing';

import { HttpClient, HttpErrorResponse } from '@angular/common/http';

import { JWTAuthService } from './jwt-auth.service';
import { JwtHelperService, JWT_OPTIONS, JwtModule } from '@auth0/angular-jwt';
import { AuthUser, AuthRequest } from '../models/auth-user.model';


// Storage
import { SecuredStorageProvider, SecurityStorageToken } from './secured-storage.provider';


// config
import { CookieService } from 'ngx-cookie-service';
import { SecureStorageService } from './secure-storage.service';
import { SecurityConfig, SecurityConfigurationService } from './security-configuration.service';
import { SecurityMockConfig } from '../mocks/mock-config';
import { MockStorageProvider } from '../../storage/mocks/mock-storage-provider';
import { AuthEvents } from './authentication.provider';
import { RouterTestingModule } from '@angular/router/testing';
import { LoginComponent } from '../components/login/login.component';

// tslint:disable-next-line:max-line-length
const token = 'eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJodHRwOi8vbG9jYWwtYXBpLnRoZWZpbmRlcnNrZWVwZXJzbWFya2V0cGxhY2UuY29tL3YxL2F1dGhlbnRpY2F0ZSIsImlhdCI6MTUyMDkxMDI1NywiZXhwIjoxNzIwOTEzODU3LCJuYmYiOjE1MjA5MTAyNTcsImp0aSI6InRnRklzak9qSVVESjFtUm0iLCJzdWIiOjEsInBydiI6IjIzYmQ1Yzg5NDlmNjAwYWRiMzllNzAxYzQwMDg3MmRiN2E1OTc2ZjcifQ.qX2sxX2MSOT - mCpvXzkVDVQURrts30xyX - vQbWTNGX0';
const loginUrl = SecurityMockConfig.apiUrl + SecurityMockConfig.apiEndpoints.authUrl;
const profileUrl = SecurityMockConfig.apiUrl + SecurityMockConfig.apiEndpoints.profileUrl('unknownUser');

describe('JWTAuthService', () => {
  let httpClient: HttpClient;
  let httpTestingController: HttpTestingController;
  const auth: AuthRequest = { user_login: 'unknownUser', password: 'unknown Password', access_token: 'token' };
  const user: AuthUser = new AuthUser(1, 'Test', 'User', 'email@address.com', '22-11-2019');
  user.user_login = 'unknownUser';

  beforeEach(() => {
    TestBed.configureTestingModule({
      imports: [
        JwtModule.forRoot({
          config: {
            tokenGetter: () => {
              return token;
            },
            headerName: 'TestHeader'
          }
        }),HttpClientTestingModule,
      RouterTestingModule.withRoutes( [{
        path: 'login', component: LoginComponent
      }])],
      providers: [
        JWTAuthService,
        JwtHelperService,
        CookieService,
        SecureStorageService,
        { provide: SecuredStorageProvider, useFactory: () => new SecureStorageService(new MockStorageProvider<AuthUser>()) },
        {
          provide: SecurityConfig,
          useValue: SecurityMockConfig
        },
        SecurityConfigurationService
      ],
      declarations: [LoginComponent]
    });
    httpClient = TestBed.get(HttpClient);
    httpTestingController = TestBed.get(HttpTestingController);

  });
  afterEach(() => {
    // After every test, assert that there are no more pending requests.
    // httpTestingController.verify();
  });

  it('should be created', inject([JWTAuthService], (service: JWTAuthService) => {
    expect(service).toBeTruthy();
  }));

  // Test
  it('should be sending authRequest', inject([JWTAuthService], (service: JWTAuthService) => {
    service.onAuthChange().subscribe(authuser => {
      // NB: with testingController - this is just a pass throug service - no mapping dome
      expect(authuser).toBeDefined();
      const newUser = new AuthUser();
      newUser.copyConstruct(authuser.authData);
      expect(newUser.getId()).toEqual(1);
      expect(newUser.firstName).toEqual('Test');
      expect(newUser.user_login).toEqual('unknownUser');
    }
    );
    service.authenticate('unknownUser', 'unknown Password').subscribe();
    // we;ll need an environment here
    const req = httpTestingController.expectOne(loginUrl);
    expect(req.request.method).toEqual('POST');
    req.flush(auth);

    const req2 = httpTestingController.expectOne(profileUrl);
    expect(req2.request.method).toEqual('GET');
    req2.flush({ data: { userprofile: user } });
  }));

  it('expects a 401 for unknown login', inject([JWTAuthService], (service: JWTAuthService) => {
    const msg = 'Error - not found';
    service.onAuthChange().subscribe(
      data => {
        // should be handled ?
        console.log( data );
      }
    );
    service.authenticate('unknownUser', 'unknown Password').subscribe( result => console.log(result), error => {
      expect(error.status).toEqual(401, 'status');
      expect(error.error).toEqual(msg, 'message');
    });
    // we;ll need an environment here
    const req = httpTestingController.expectOne(loginUrl);
    expect(req.request.method).toEqual('POST');
    req.flush(msg, { status: 401, statusText: 'Not Found' });
  }));

  it('should store token after successful request - but not authenticated', inject([JWTAuthService], (service: JWTAuthService) => {
    // tslint:disable-next-line:max-line-length
    const err: Error = new Error('The inspected token doesn\'t appear to be a JWT.Check to make sure it has three parts and see https://jwt.io for more');
    // service.logout();
    service.onAuthChange().subscribe(authuser => {
      // expect(service.getAuthToken()).toEqual('token');
      expect(service.isAuthenticated).toThrow();
      // should fail authentication becuase no expiry
    }
    );
    service.authenticate('unknownUser', 'unknwon Password').subscribe();
    // Auth Code
    const req = httpTestingController.expectOne(loginUrl);
    expect(req.request.method).toEqual('POST');
    req.flush(auth);

    const req2 = httpTestingController.expectOne(profileUrl);
    expect(req2.request.method).toEqual('GET');
    req2.flush({ data: { userprofile: user } });
  }));

  it('should store valid token after successful request -  authenticated', inject([JWTAuthService], (service: JWTAuthService) => {
    // tslint:disable-next-line:max-line-length
    const token = 'eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJodHRwOi8vbG9jYWwtYXBpLnRoZWZpbmRlcnNrZWVwZXJzbWFya2V0cGxhY2UuY29tL3YxL2F1dGhlbnRpY2F0ZSIsImlhdCI6MTUyMDkxMDI1NywiZXhwIjoxNzIwOTEzODU3LCJuYmYiOjE1MjA5MTAyNTcsImp0aSI6InRnRklzak9qSVVESjFtUm0iLCJzdWIiOjEsInBydiI6IjIzYmQ1Yzg5NDlmNjAwYWRiMzllNzAxYzQwMDg3MmRiN2E1OTc2ZjcifQ.qX2sxX2MSOT - mCpvXzkVDVQURrts30xyX - vQbWTNGX0';
    auth.access_token = token;
    // tslint:disable-next-line:max-line-length
    const err: Error = new Error('The inspected token doesn\'t appear to be a JWT.Check to make sure it has three parts and see https://jwt.io for more');

    service.onAuthChange().subscribe(authuser => {
      // expect(authuser).toEqual(user);
      expect(service.getAuthToken()).toEqual(token);
      expect(service.isAuthenticated()).toBeTruthy();
    }
    );
    service.authenticate('unknownUser', 'unknwon Password').subscribe();
    const req = httpTestingController.expectOne(loginUrl);
    expect(req.request.method).toEqual('POST');
    req.flush(auth);

    const req2 = httpTestingController.expectOne(profileUrl);
    expect(req2.request.method).toEqual('GET');
    req2.flush({ data: { userprofile: user } });

  }));

  it('should fail role checks on empty roles', inject([JWTAuthService], (service: JWTAuthService) => {
    // tslint:disable-next-line:max-line-length
    const token = 'eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJodHRwOi8vbG9jYWwtYXBpLnRoZWZpbmRlcnNrZWVwZXJzbWFya2V0cGxhY2UuY29tL3YxL2F1dGhlbnRpY2F0ZSIsImlhdCI6MTUyMDkxMDI1NywiZXhwIjoxNzIwOTEzODU3LCJuYmYiOjE1MjA5MTAyNTcsImp0aSI6InRnRklzak9qSVVESjFtUm0iLCJzdWIiOjEsInBydiI6IjIzYmQ1Yzg5NDlmNjAwYWRiMzllNzAxYzQwMDg3MmRiN2E1OTc2ZjcifQ.qX2sxX2MSOT - mCpvXzkVDVQURrts30xyX - vQbWTNGX0';
    auth.access_token = token;
    // tslint:disable-next-line:max-line-length
    const err: Error = new Error('The inspected token doesn\'t appear to be a JWT.Check to make sure it has three parts and see https://jwt.io for more');

    service.onAuthChange().subscribe(authuser => {
      expect(service.getAuthToken()).toEqual(token);
      expect(service.isAuthenticated()).toBeTruthy();
      expect(service.hasRole('any')).toBeFalsy();
      expect(service.checkRoles(['role1', 'role2'])).toBeFalsy();
      // should fail authentication becuase no expiry
    }
    );
    service.authenticate('unknownUser', 'unknwon Password').subscribe()
    const req = httpTestingController.expectOne(loginUrl);
    expect(req.request.method).toEqual('POST');
    req.flush(auth);

    const req2 = httpTestingController.expectOne(profileUrl);
    expect(req2.request.method).toEqual('GET');
    req2.flush({ data: { userprofile: user } });

  }));

  it('should pass appropriate role checks', inject([JWTAuthService], (service: JWTAuthService) => {
    // tslint:disable-next-line:max-line-length
    const token = 'eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJodHRwOi8vbG9jYWwtYXBpLnRoZWZpbmRlcnNrZWVwZXJzbWFya2V0cGxhY2UuY29tL3YxL2F1dGhlbnRpY2F0ZSIsImlhdCI6MTUyMDkxMDI1NywiZXhwIjoxNzIwOTEzODU3LCJuYmYiOjE1MjA5MTAyNTcsImp0aSI6InRnRklzak9qSVVESjFtUm0iLCJzdWIiOjEsInBydiI6IjIzYmQ1Yzg5NDlmNjAwYWRiMzllNzAxYzQwMDg3MmRiN2E1OTc2ZjcifQ.qX2sxX2MSOT - mCpvXzkVDVQURrts30xyX - vQbWTNGX0';
    auth.access_token = token;

    // tslint:disable-next-line:max-line-length
    const err: Error = new Error('The inspected token doesn\'t appear to be a JWT.Check to make sure it has three parts and see https://jwt.io for more');

    service.onAuthChange().subscribe(authuser => {
      expect(service.getAuthToken()).toEqual(token);
      expect(service.isAuthenticated()).toBeTruthy();
      // NB: Maybe a better unittest is to use the secured storage provider.
      // This requires TestBackend to be in place - really an integration test
      // expect(service.hasRole('role2')).toBeFalsy();
      // expect(service.hasRole('role1')).toBeTruthy();
      // expect(service.checkRoles(['role1', 'role2'])).toBeTruthy();
      // should faile authentication becuase no expiry
    }
    );
    service.authenticate('unknownUser', 'unknwon Password').subscribe(result => console.log(result), error => console.log(error));
    // Auth Code
    const req = httpTestingController.expectOne(loginUrl);
    expect(req.request.method).toEqual('POST');
    req.flush(auth);

    const req2 = httpTestingController.expectOne(profileUrl);
    expect(req2.request.method).toEqual('GET');
    req2.flush({ data: { userprofile: user } });
  }));

  it('should logout by deleting storage and sending request', inject([JWTAuthService], (service: JWTAuthService) => {
    // tslint:disable-next-line:max-line-length
    const token = 'eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJodHRwOi8vbG9jYWwtYXBpLnRoZWZpbmRlcnNrZWVwZXJzbWFya2V0cGxhY2UuY29tL3YxL2F1dGhlbnRpY2F0ZSIsImlhdCI6MTUyMDkxMDI1NywiZXhwIjoxNzIwOTEzODU3LCJuYmYiOjE1MjA5MTAyNTcsImp0aSI6InRnRklzak9qSVVESjFtUm0iLCJzdWIiOjEsInBydiI6IjIzYmQ1Yzg5NDlmNjAwYWRiMzllNzAxYzQwMDg3MmRiN2E1OTc2ZjcifQ.qX2sxX2MSOT - mCpvXzkVDVQURrts30xyX - vQbWTNGX0';
    auth.access_token = token;

    // tslint:disable-next-line:max-line-length
    let loginIndex = 0;
    let i = 0;
    service.onAuthChange().subscribe(authuser => {
      console.log(authuser);
      if( authuser.eventType === AuthEvents.login) {
        expect(service.isAuthenticated()).toBeTruthy();
        loginIndex = i;
      }
      if( authuser.eventType === AuthEvents.logout) {
        expect(service.isAuthenticated()).toBeFalsy();
        expect(service.getAuthUser()).toBeNull();
        expect(i).toBeGreaterThan(loginIndex);
        // Logout should follow login!
      }
      i++;
    });

    service.authenticate('unknownUser', 'unknown Password').subscribe( result => service.logout());
    
    // Auth Code
    const req = httpTestingController.expectOne(loginUrl);
    expect(req.request.method).toEqual('POST');
    req.flush(auth);

    const req2 = httpTestingController.expectOne(profileUrl);
    expect(req2.request.method).toEqual('GET');
    req2.flush({ data: { userprofile: user } });

  }));
});
