/**
 * JWT AuthService
 * Copyritght: the Finders Keepers
 * Created: 20/3/2018
 */

import { Injectable } from '@angular/core';
import { HttpClient, HttpHeaders } from '@angular/common/http';

// Rxjs
import { Observable, BehaviorSubject, Subject } from 'rxjs';
import { map, switchMap, take } from 'rxjs/operators';

// JWT
import { JwtHelperService } from '@auth0/angular-jwt';

// Models
import { AuthUser, AuthRequest } from '../models/auth-user.model';


// Storage
import { AuthenticationProvider, AuthenticationEvent, AuthEvents } from './authentication.provider';
import { SecurityConfigurationService } from './security-configuration.service';
import { IAuthenticationConfiguration } from '../models/auth-configuration';
import { SecuredStorageProvider } from './secured-storage.provider';
import { isEmpty } from '../../data/functions/is-empty';


/**
 * Export
 * An Authentication service that relies on a JWT to
 * handle security
 */
@Injectable()
export class JWTAuthService extends AuthenticationProvider {

  /**
   * Private  of jwtauth service
   * restful URLS for authentication provides
   */
  private serviceAuthUrl: string;
  private config: IAuthenticationConfiguration;
  private authState: Subject<boolean> = new Subject<boolean>();
  public authEventState = new Subject<AuthenticationEvent>();

  /**
   * Http options of jwtauth service
   */
  private httpOptions: { headers: HttpHeaders } = { headers: new HttpHeaders({ 'Content-Type': 'application/json'}) };

  private _tokenServiceUrl: string;
  /**
   * Creates an instance of jwtauth service.
   * parameter: jwtHelper
   * parameter: http
   * parameter: config
   * parameter: storageProvider
   */
  constructor(private jwtHelper: JwtHelperService, private http: HttpClient, config: SecurityConfigurationService,
              private storageProvider: SecuredStorageProvider) {
    super();
    this.setUrl( config.getConfiguration() );
    this.storageProvider.setDomain(config.getConfiguration().securedDomain);
    this.config = config.getConfiguration();
  }

  /**
   * Sets url
   * parameter: config
   */
  private setUrl(config: IAuthenticationConfiguration) {
    this.serviceAuthUrl = config.apiUrl + config.apiEndpoints.authUrl;
    // Configuration needs to perform translation
  }

  /**
   * Checks the existence and validity
   * of a JWT
   * Returns jwtauth service
   * @returns true if authenticated
   */
  public isAuthenticated(): boolean {
    const token = this.storageProvider.getAuthToken();
    if( isEmpty( token ) ) {
      return false;
    }
    // Check whether the token is expired and return
    try {
      // Any other work on token
      return !this.jwtHelper.isTokenExpired(token);
    } catch (e) {
      console.log( 'fails ');
      throw(e);
    }

  }
  /**
   * Authenticates jwtauth service
   * userName
   * password
   * authenticate
   */
  public authenticate(userName: string, password: string): Observable<AuthUser> {
    this.clearSession();
    // return this.http.get<Auth>( this.serviceAuthUrl);
    // switchMap to ensure subscriber is passed on but not yet modular
    // Somehow the token needs to be generic
    const authRequest: AuthRequest = new AuthRequest();
    authRequest.user_login = userName;
    authRequest.password = password;
    return this.http.post<AuthRequest>(this.serviceAuthUrl, authRequest, this.httpOptions).pipe(switchMap( (auth, index: number) => {
      // const id = auth.userID;
      const token = auth.access_token;
      this.storageProvider.setAuthToken(token);
      // @TODO Set Username
      // console.info('Signed In');
      return this.http.get<{ data: { userprofile: AuthUser } }>(this.getProfileUrl(authRequest.user_login)).pipe( map ( result => {
        this.storageProvider.setAuthUser(result.data.userprofile);
        // @TODO Set Local Culture
        this.authState.next( true );
        this.authEventState.next( {
          eventType: AuthEvents.login,
          authData: result.data.userprofile,
          rawData: result
        });
        return result.data.userprofile;
      }, error => this.authEventState.next( {
        eventType: AuthEvents.failedLogin,
        authData: null,
        rawData: error})
      )) // Only take the first response to close of the subscription
      // setUSer
    }));
  }
  private getProfileUrl(username: string): string {
    return this.config.apiUrl + this.config.apiEndpoints.profileUrl(username);
  }
  /**
   * Logouts jwtauth service -
   * Creates Observable
   * @returns true if logout
   */
  public logout(): boolean {
    this.clearSession();
    // check cleared
    this.authState.next( false );
    this.authEventState.next( {
      authData: undefined,
      eventType: AuthEvents.logout,
      rawData: undefined
    })
    return !this.isAuthenticated();
  }
  private clearSession(): boolean {
    this.storageProvider.deleteStorage();
    return true;
  }

  /**
   * Gets auth user from session storage
   * @returns auth user
   */
  public getAuthUser(): AuthUser {
    return this.isAuthenticated() ? this.storageProvider.getAuthUser() : null;
  }

  /**
   * Gets raw auth token
   * @returns auth token
   */
  public getAuthToken(): string {
    return this.storageProvider.getAuthToken();
  }

  /**
   * Send an array of roles and check if user is in one of them
   * parameter: expectedRoles
   * @returns true if roles
   */
  public checkRoles(expectedRoles: string[]): boolean {
    let allowUser = false;
    expectedRoles.forEach(role => {
      // this.storageProvider.getRoles()
      if ( this.hasRole(role) ) {
        allowUser = true;
      }
    });
    return allowUser;
  }
  /**
   * Determines whether user has role
   * parameter: role
   * @returns true if role
   */
  public hasRole(role: string): boolean {
    const user = this.getAuthUser();
    if ( user !== null && user.user_access_name !== undefined) {
      return this.storageProvider.getAuthUser().user_access_name.toLowerCase() === role.toLowerCase();
    }
    return false;
  }
  /**
   * Gest the auth change subscription
   * @returns auth change
   */
  public onAuthChange(): Observable<AuthenticationEvent> {
    return this.authEventState;
  }
  /**
   * Refreshes session
   */
  public refreshSession(): void {
    // Needs to be a reset endpoint
    throw new Error('Method not implemented.');
  }
  public getAuthenticatedState(): Observable<boolean> {
    return this.authState.asObservable();
  }
}

