// Core
import { Injectable } from '@angular/core';

// Local Imports
import { SecuredStorageProvider, StorageKeys, SecurityStorageToken} from './secured-storage.provider';
import { AuthUser } from '../models/auth-user.model';
import { StorageProvider } from '../../storage/services/storage.provider';


/**
 * The secure storage provider has a storage provider
 * injected. Arguably the abstraction is not needed.
 */
@Injectable()
export class SecureStorageService extends SecuredStorageProvider {

  /**
   * Creates an instance of secure storage service.
   * @param storage Storage Provider from a factory
   */
  constructor(private storage: StorageProvider<AuthUser>) {
    super();
   }

  /**
   * Set the JWT in a secured cookie, default to local storage
   * @param token JWT
   */
  public setAuthToken(token: string) {
    this.storage.deleteKey(StorageKeys[StorageKeys.authToken]);
    this.storage.setKeyValue(StorageKeys[StorageKeys.authToken], token);
  }

  /**
   * Gets auth token
   * @returns auth token
   */
  public getAuthToken(): string {
    return this.storage.getKeyValue(StorageKeys[StorageKeys.authToken]);
  }

  /**
   * Deletes storage
   */
  public deleteStorage() {
    this.storage.clearData();
  }

  /**
   * Sets domain
   * @param domain secure domain with protocol
   */
  public setDomain(domain: string) {
    this.storage.setDomain(domain);
  }

  /**
   * Sets auth user
   * @param user AuthUser Model
   */
  public setAuthUser(user: AuthUser): void {
    this.storage.deleteKey(StorageKeys[StorageKeys.authUser]);
    this.storage.setObjectValue(StorageKeys[StorageKeys.authUser], user);
  }

  /**
   * Gets auth user
   * @returns auth user
   */
  public getAuthUser(): AuthUser {
    const user =  new AuthUser();
    user.copyConstruct(this.storage.getKeyObjectValue(StorageKeys[StorageKeys.authUser]));
    return user;
  }
}
