import { TestBed, inject } from '@angular/core/testing';

import { HttpClientTestingModule, HttpTestingController } from '@angular/common/http/testing';
import { HttpClient } from '@angular/common/http';
import { IResponseModel } from '../models/response-model';
import { handleCloneableResponse, handleCloneableListResponse } from './constructed-response-handler.service';
import { HelloWorld, MockModel } from '../mocks/mock-cloneable';
import { MessageProvider, MessageProviderToken } from '../../messaging/services/message-provider';
import { MessagingModule } from '../../messaging/messaging.module';
import { MessagingService } from '../../messaging/services/message.service';
import { ResponseHandlerService } from './response-handler.service';
import { MessageSeverity } from '../../messaging/models/i-message';


const model: HelloWorld = {
    value: 'hello',
    name: 'world'
};
const listModel: HelloWorld[] = [model, model, model]; 
const listResponse: IResponseModel<HelloWorld[]> = {
  data: listModel,
  code: '200',
  messsage: ''
};

const response: IResponseModel<HelloWorld> = {
  data: model,
  code: '200',
  messsage: ''
};
const mockModel = new MockModel();
mockModel.copyConstruct(model);
const responseMockModel: IResponseModel<MockModel> = {
  data: mockModel,
  code: '200',
  messsage: ''
};

const responseListMockModel: IResponseModel<MockModel[]> = {
  data: [mockModel, mockModel, mockModel],
  code: '200',
  messsage: ''
};

const simpleResponse: IResponseModel<boolean> = {
  data: true,
  code: '200',
  messsage: ''
};
describe('ResponseHandlerService', () => {
  let httpMock: HttpTestingController;
  let messageservice: MessageProvider;
  let httpClient: HttpClient;
  beforeEach(() => {
    TestBed.configureTestingModule({
    imports: [HttpClientTestingModule, MessagingModule],
    providers: [
        {
          provide: MessageProviderToken,
          useClass: MessagingService
        },
        // NB: Whilst we factory inject for testing - the expected usage is to extend the handler
        // NB2: The instantiation of the class Type - important as no easy access to reflexion in typescript
        {
          provide: 'httpHandler',
          deps: [HttpClient, MessageProviderToken],
          useFactory: (client: HttpClient, messager: MessageProvider) => new ResponseHandlerService<HelloWorld>(client, messager)
        }]
    });
    messageservice = TestBed.get(MessageProviderToken)
    httpClient = TestBed.get(HttpClient);
    httpMock = TestBed.get(HttpTestingController);
  });
  it('should be created', inject( ['httpHandler'], (service: ResponseHandlerService<HelloWorld>) => {
    expect(service).toBeTruthy();
  }));
  it('should handle GET request', inject( ['httpHandler'], (service: ResponseHandlerService<HelloWorld>) => {
    let iterator = 0;
    messageservice.getMessages().subscribe( messages => {
      switch (iterator) {
        case 1:
          expect(messages.length).toEqual(1);
          expect(messages[0].severity).toEqual(MessageSeverity.info);
          expect(messages[0].summary).toEqual('200');
          break;
      }
      iterator ++;
    })
    service.handleGet('/someUrl').subscribe( result => {
      expect(result).toEqual(model);
    });
    const testRequest = httpMock.expectOne('/someUrl');
    expect(testRequest.request.method).toEqual('GET');
    testRequest.flush(response);
    httpMock.verify();
  }));
  it('should handle POST request', inject( ['httpHandler'], (service: ResponseHandlerService<HelloWorld>) => {
    let iterator = 0;
    messageservice.getMessages().subscribe( messages => {
      switch (iterator) {
        case 1:
          expect(messages.length).toEqual(1);
          expect(messages[0].severity).toEqual(MessageSeverity.info);
          expect(messages[0].summary).toEqual('200');
          break;
      }
      iterator ++;
    });
    service.handlePost('/someUrl', model).subscribe( result => {
      expect(result).toEqual(model);
      // Cannot mock this - but shallow copy would normally occure expect(result.getText).toBeUndefined();
    });
    const testRequest = httpMock.expectOne('/someUrl');
    expect(testRequest.request.method).toEqual('POST');
    expect(testRequest.request.body).toEqual(model);
    testRequest.flush(response);
    httpMock.verify();
  }));
  it('should handle PUT request', inject( ['httpHandler'], (service: ResponseHandlerService<HelloWorld>) => {
    let iterator = 0;
    messageservice.getMessages().subscribe( messages => {
      switch (iterator){
        case 1: 
          expect(messages.length).toEqual(1);
          expect(messages[0].severity).toEqual(MessageSeverity.info);
          expect(messages[0].summary).toEqual('200');
          break;
      }
      iterator ++;
    })
    service.handlePut('/someUrl', model).subscribe( result => {
      expect(result).toEqual(model);
    });
    const testRequest = httpMock.expectOne('/someUrl');
    expect(testRequest.request.method).toEqual('PUT');
    expect(testRequest.request.body).toEqual(model);
    testRequest.flush(response);
    httpMock.verify();
  }));
  it('should handle DELETE request', inject( ['httpHandler'], (service: ResponseHandlerService<HelloWorld>) => {
    let iterator = 0;
    messageservice.getMessages().subscribe( messages => {
      switch (iterator){
        case 1: 
          expect(messages.length).toEqual(1);
          expect(messages[0].severity).toEqual(MessageSeverity.info);
          expect(messages[0].summary).toEqual('200');
          break;
      }
      iterator ++;
    })
    service.handleDelete('/someUrl').subscribe( result => {
      expect(result).toBeTruthy();
    });
    const testRequest = httpMock.expectOne('/someUrl');
    expect(testRequest.request.method).toEqual('DELETE');
    testRequest.flush(simpleResponse);
    httpMock.verify();
  }));
  it('should handle List of T request', inject( ['httpHandler'], (service: ResponseHandlerService<HelloWorld>) => {
    let iterator = 0;
    messageservice.getMessages().subscribe( messages => {
      switch (iterator){
        case 1: 
          expect(messages.length).toEqual(1);
          expect(messages[0].severity).toEqual(MessageSeverity.info);
          expect(messages[0].summary).toEqual('200');
          break;
      }
      iterator ++;
    })
    service.handleGetList('/someUrl').subscribe( result => {
      expect(result).toEqual(listModel);
    });
    const testRequest = httpMock.expectOne('/someUrl');
    expect(testRequest.request.method).toEqual('GET');
    testRequest.flush(listResponse);
    httpMock.verify();
  }));
  it('should set construct handler for GET request', inject( ['httpHandler'], (service: ResponseHandlerService<MockModel>) => {
    service.setHandler(handleCloneableResponse);
    expect(service.handlerSet('handleCloneableResponse')).toBeTruthy();
    let iterator = 0;
    messageservice.getMessages().subscribe( messages => {
      iterator ++;
      // no messages sent when we change handler!
      expect(iterator).toEqual(1);
    });
    service.handleGet('/someUrl').subscribe( result => {
      expect(result).toEqual(mockModel);
      expect(result.getText()).toEqual('hello world');
    });
    const testRequest = httpMock.expectOne('/someUrl');
    expect(testRequest.request.method).toEqual('GET');
    testRequest.flush(responseMockModel);
    httpMock.verify();
  }));
  it('should set construct handler for GET List request', inject( ['httpHandler'], (service: ResponseHandlerService<MockModel>) => {
    service.setListHandler(handleCloneableListResponse, MockModel);
    let iterator = 0;
    messageservice.getMessages().subscribe( messages => {
      iterator ++;
      // no messages sent when we change handler!
      expect(iterator).toEqual(1);
    });
    service.handleGetList('/someUrl').subscribe( result => {
      expect(result).toEqual([mockModel, mockModel, mockModel]);
      result.forEach( modeled => expect(modeled.getText()).toEqual('hello world'));
    });
    const testRequest = httpMock.expectOne('/someUrl');
    expect(testRequest.request.method).toEqual('GET');
    testRequest.flush(responseListMockModel);
    httpMock.verify();
  }));
  it('should set and remove handler', inject( ['httpHandler'], (service: ResponseHandlerService<MockModel>) => {
    service.setHandler(handleCloneableResponse);
    expect(service.handlerSet('handleCloneableResponse')).toBeTruthy();
    service.removeHandler();
    expect(service.handlerSet('handleCloneableResponse')).toBeFalsy();
    expect(service.handlerSet('')).toBeFalsy();

    service.setListHandler(handleCloneableListResponse, MockModel);
    expect(service.listHandlerSet('handleCloneableListResponse')).toBeTruthy();
    service.removeListHandler();
    expect(service.listHandlerSet('handleCloneableListResponse')).toBeFalsy();
    // expect(service.listHandlerSet('')).toBeTruthy();
  }));
});
