import { Injectable, Inject } from '@angular/core';
import { MetadataResponseModel, IResponseModel } from '../models/response-model';
import { HttpClient } from '@angular/common/http';
import { map } from 'rxjs/operators';
import { Observable } from 'rxjs';
import { MessageProviderToken, MessageProvider } from '../../messaging/services/message-provider';

export type genericHandler<T> = (response: IResponseModel<T>, type?: new () => T) => T;
export type genericListHandler<T> = (response: IResponseModel<T[]>, type?: new () => T) => T[];
/**
 * Injectable Handler
 * NB: It occurs to me that I could write this as an interceptor. But that could get painful to remove.
 * Currently we expect a user to extend this class with a concrete Type and add local Handlers as required
 * NB2: You would extend this class to string requests together via mergeMap/swicthMap etc.
 * template T
 */
@Injectable({
  providedIn: 'root'
})
export class ResponseHandlerService<T> {

  private listHandler: genericListHandler<T>;
  private handler: genericHandler<T>;
  
  private type: (new () => T);
  constructor(private _http: HttpClient,
              @Inject(MessageProviderToken) private provider: MessageProvider) {

    this.handler = this.handleResponse;
    this.listHandler = this.handleListResponse;
  }

  /**
   * Sets type
   *  type
   */
  public setType(type: (new () => T)) {
    this.type = type;
  }

  /**
   * Handles get
   * url
   * get
   */
  public handleGet(url: string): Observable<T> {
    return this._http.get<IResponseModel<T>>(url).pipe( map(
      result => this.handler(result, this.type)
    ));

  }

  /**
   * Sets list handler
   *  handler
   *  type
   */
  public setListHandler(handler: genericListHandler<T>, type: (new () => T)): void {
    this.setType(type);
    this.listHandler = handler;
  }
  /**
   * Sets handler
   * NB: The type was originally set in the constructor -
   *     but this will fail on ngPackagr because dependency injection is not set up
   *  handler
   *  type
   */
  public setHandler(handler: genericHandler<T>): void {
    
    this.handler = handler;
  }

  /**
   * Handlers set
   *  handlerName
   * returns true if set
   */
  public handlerSet(handlerName: string): boolean {
    return this.handler.name === handlerName;
  }

  /**
   * Lists handler set
   *  handlerName
   * returns true if handler set
   */
  public listHandlerSet(handlerName: string): boolean {
    return this.listHandler.name === handlerName;
  }

  /**
   * Removes list handler
   */
  public removeListHandler(): void {
    this.listHandler = this.handleListResponse;
  }
  /**
   * Removes handler
   */
  public removeHandler(): void {
    this.handler = this.handleResponse;
  }

  /**
   * Handles get list
   *  url
   * returns get list
   */
  public handleGetList(url: string): Observable<T[]> {
    return this._http.get<IResponseModel<T[]>>(url).pipe(map(
      result => this.listHandler(result, this.type)
    ));
  }

  /**
   * Handles post
   *  url
   *  data
   * returns post
   */
  public handlePost(url: string, data: T): Observable<T> {
    return this._http.post<IResponseModel<T>>(url, data).pipe(map(
      result => this.handler(result, this.type)
    ));
  }
  /**
   * Handles post
   *  url
   *  data
   * post
   */
  public handlePostTransform<Input>(url: string, data: Input): Observable<T> {
    return this._http.post<IResponseModel<T>>(url, data).pipe(map(
      result => this.handler(result, this.type)
    ));
  }
  /**
   * Handles server side filter
   * template FilterType
   * param url
   * param data
   * returns server side filter
   */
  public handleServerSideFilter<FilterType>(url: string, data: FilterType): Observable<T[]> {
    return this._http.post<IResponseModel<T[]>>(url, data).pipe(map(
      result => this.listHandler(result, this.type)
    ));
  }

  /**
   *
   * param url
   * param data
   */
  public handleServerSideFilterWithMetadata<FilterType, Metadata>(url: string, data: FilterType):
       Observable<MetadataResponseModel<T[], Metadata>> {
    return this._http.post<MetadataResponseModel<T[], Metadata>>(url, data);
  }
  /**
   * Handles delete
   * url
   * delete
   */
  public handleDelete(url: string): Observable<boolean> {
    return this._http.delete<IResponseModel<boolean>>(url).pipe(map(
      result => result.data
    ));
  }

  /**
   * Handles put
   *  url
   *  data
   * returns put
   */
  public handlePut(url: string, data: T): Observable<T> {
    return this._http.put<IResponseModel<T>>(url, data).pipe(map(
      result => this.handler(result, this.type)
    ));
  }
  /**
   * Handles response - dropped the templating
   *  response
   * returns response
   */
  protected handleResponse(response: IResponseModel<T>): T {
    this.provider.sendInfoMessage(response.code, response.messsage);
    return response.data;
  }
  /**
   * Handles list response
   * Requires to versions to allow for better extension
   * template T dropped
   *  response
   * returns list response
   */
  protected handleListResponse(response: IResponseModel<T[]>): T[] {
    this.provider.sendInfoMessage(response.code, response.messsage);
    return response.data;
  }
}
