import { IMenu, IMenuItem } from "./i-menu";

export enum NavigationHistory {
    current = 0,
    previous = 1,
    forward = 2    
}
export class Menu implements IMenu {
    private _routes: IMenuItem[] = [];
    private readonly ROOT_TEXT = "10dd9f47-f13f-4865-9c8a-40e45ac68c26";
    private _navigation:Map<NavigationHistory, IMenuItem> = new Map<NavigationHistory, IMenuItem>();
    // This is used for a select multiple parent nodes call
    private _parentChild: Map<number, number[]> = new Map<number, number[]>();
    constructor() {
        const root: IMenuItem = {
            id: 0,
            title: this.ROOT_TEXT,
            url: this.ROOT_TEXT,
            path: '',
            parentId: -1,
            isTopLevel: false
        }
        this._routes.push(root);
        this._parentChild.set(0, []);
    }
    public generateUrls(baseUrl = ""): void {
        this._routes.forEach( item => item.url = this.generateUrl(item, baseUrl));
    }
    /**
     * Adds route
     * @param item 
     */
    public addRoute(item: IMenuItem, baseUrl= "") {
        if(item.id === 0 ){
            throw("Can not add new root items");
        }
        // in case of existing we will just skip... could throw
        if( this.findItem(item.id) !== undefined) {
            return;
        }
        //urls musty be unique - which means we may need to join the path
        
        if( this._parentChild.has(item.parentId) ){
            const items = this._parentChild.get(item.parentId);
            items.push(item.id);
            this._parentChild.set(item.parentId, items);           
        } else {
            this._parentChild.set(item.parentId, [item.id]);            
        }    
        item.url = this.generateUrl(item, baseUrl);
        // uniqueness must occur on ID instead of URL as we
        // can double up on paths due to parent/child having the same path
        // but pointing to different location
        if( this.findItem(item.id) !== undefined) {
            return;
        }
        this._routes.push(item);
            
    }
    public generateUrl(item: IMenuItem, baseUrl = ""): string {
        const fullPath = this.getPathway(item, this._routes.length);
        let url = baseUrl;
        fullPath.forEach( subpath => {
            subpath.path.length !== 0 ? 
                url = url + '/' + subpath.path :
                url = url + subpath.path
        })
        return url;
    }
    /**
     * Gets active route
     * @returns active route 
     */
    public getActiveRoute(): IMenuItem {
        return this._navigation.get(NavigationHistory.current);
    }    

    /**
     * Sets active path
     * @param url 
     */
    public setActivePath(url: string): void {
        const item = this.findByPath(url);
        if( item !== undefined){
            const prev = this._navigation.get(NavigationHistory.current);
            this._navigation.set(NavigationHistory.previous, prev);
            this._navigation.set(NavigationHistory.current, item);
        }
    }  
    /**
     * 
     * @param id 
     */  
    public getChildrenForMenuItem(id: number): IMenuItem[] {
        return this._routes.filter(items => items.parentId === id);
    }
    /**
     * Finds item
     * @param id 
     * @returns item 
     */
    public findItem(id: number): IMenuItem {
        return this._routes.find( x => x.id === id);
    }
    /**
     * Finds by path
     * @param path 
     * @returns by path 
     */
    public findByPath(path: string): IMenuItem {
        return this._routes.find( x => x.url === path);
    }
    /**
     * Gets all routes
     * @returns all routes 
     */
    public getAllRoutes(): IMenuItem[] {
        return this._routes;
    }
    /**
     * Gets all children
     * @param parentIds 
     * @returns all children 
     */
    public getAllChildren(parentIds: number[]): IMenuItem[] {
        let ids: number[] = [];
        
        parentIds.forEach( id =>
            ids = ids.concat( this._parentChild.get(id )));
        return this._routes.filter(x => ids.some( y => y === x.id));

    }
    /**
     * Gets pathway in order of root to leaf
     * @param initial 
     * @param [maxIterations] 
     * @returns pathway 
     */
    public getPathway(initial: IMenuItem, maxIterations=3, iteration = 0, breadcrumbs: IMenuItem[] = []) : IMenuItem[] {
        if( iteration === maxIterations) {
            return breadcrumbs;
        }
        if( initial === undefined ) {
            return breadcrumbs;
        }
        breadcrumbs.unshift(initial);
        
        if( this.notNullOrZero(this.findItem(initial.parentId)) ) {            
            breadcrumbs = this.getPathway(this.findItem(initial.parentId), maxIterations, iteration+1, breadcrumbs );                            
        }
        return breadcrumbs;
    }

    public getParentChildRelationships(): Map<number, number[]> {
        return this._parentChild;
    }
    /**
     * Gets previous route - equivalent to location.history(-1)
     * @returns previous route 
     */
    public getPreviousRoute(): IMenuItem {
        return this._navigation.get(NavigationHistory.previous);
    }

    /**
     * Gets depth via recursion
     * @param initial 
     * @param [depth] 
     * @returns depth 
     */
    public getDepth(initial: IMenuItem, depth = 0) : number {
        // We need to preserve a root item
        if( this.notNullOrZero(this.findItem(initial.parentId)) ) {
            depth = this.getDepth(this.findItem(initial.parentId), depth) + 1;                
        }
        return depth;
    }
    
    /**
     * Gets max depth
     * We need to traverse all nodes - so probably better if we optimise on ID instead of objects
     * Recursion is faster than iteration so whilst the following works we will rewrite: 
     * Math.max.apply(Math, this._routes.map((o) => { return this.getDepth(o); }))
     * @param [currentNodeId] 
     * @param [depth] 
     * @returns max depth 
     */
    public getMaxDepth(currentNodeId = 0, depth = 0 ) : number {
        // At the leaf
        if(this._parentChild.get(currentNodeId) === undefined) {
            return depth;            
        }
        // pre-increment, to get offset of 1 as only the root item is level 0
        depth ++;
        const result: number[] = [];
        this._parentChild.get(currentNodeId).forEach( childId => {
            result.push(this.getMaxDepth(childId, depth));            
        });

        return Math.max(...result);
        
    }
    /**
     * Nots null or zero
     * This is handy - needs more global support
     * @param object 
     * @returns true if null or zero 
     */
    private notNullOrZero(object: any): boolean {
        return object !== undefined && object!== 0;
    }
    // Not needed here - but Checking for key not set, value is 0 or undefined would also be handy
    // Especially as behavior can be unified
    // private notNullOrZero(object: any, key: string): boolean {
    //     return object !== undefined
    // }
    
}