import { Injectable, InjectionToken, Inject } from '@angular/core';
import { Route, Routes } from '@angular/router';
import { IMenuItem } from '../models/i-menu';
import { Menu } from '../models/menu';
import { NavigationProvider } from './navigation-provider';


// Tokens
export const BASE_ROUTE = new InjectionToken<string>('BASE_ROUTES');
export const ROUTES = new InjectionToken<Routes>('RLB_ROUTES');
/**
 * This should load the set of routes from routing, build a model from each routes associated data
 * and discover where the current active route amongst that model
 * It should also provide the essentials required for any menu service.
 * 
 * NB: Should be singleton !important
 * 
 * Route data model
 * @example 
 * data: { menuName: 'home', icon: 'fa-home', expectedRoles: ['roleName'] }
 */
@Injectable()
export class NavigationService extends NavigationProvider {

  /**
   * Menu  of navigation service
   */
  private _menu: Menu = new Menu();

  /**
   * Creates an instance of navigation service.
   * @param router 
   */
  public constructor(@Inject(ROUTES) routes: Routes, @Inject(BASE_ROUTE) baseUrl = '') {
    super();
    // Consider putting this in local storage - it can not change when built from config file
    // Routes has to be a multi-provider
    if (routes instanceof Array) {
      const resolve = routes[0];
      if (resolve instanceof Array) {
        this.buildMenu(resolve);
      }
      // May not be inserted in tree node -> leaf order.
    } else {
      this.buildMenu(routes);
    }
    // May not be inserted in tree node -> leaf order.
    this._menu.generateUrls(baseUrl);
  }

  /**
   * Builds menu - 
   * NB: route.data requires: data: {id: number, title: string, icon: string, expectedRoles: string[], parentId: number }
   * Builds a flat menu.
   * @param input 
   * @returns menu 
   */
  private buildMenu(input: Route[]): void {
    if (input === undefined) {
      return;
    }
    input.forEach(route => {
      if (route.data === undefined) {
        return;
      }
      if (!('id' in route.data)) {
        throw new Error('An ID must be set in route.data');
      }
      if (route.data !== undefined) {
        const item: IMenuItem = {
          id: route.data.id,
          title: 'title' in route.data ? route.data.title : '',
          path: route.path,
          icon: 'icon' in route.data ? route.data.icon : '',
          children: [],
          roles: 'expectedRoles' in route.data ? route.data.expectedRoles : [],
          parentId: 'parentId' in route.data ? route.data.parentId : 0,
          isTopLevel: 'isTopLevel' in route.data ? route.data.isTopLevel : false,
        };
        this._menu.addRoute(item);
        this.buildMenu(route.children)
      }
    });
  }
  /**
   * Gets active route
   * @returns active route
   */
  public setActiveRoute(url: string): void {
    return this._menu.setActivePath(url);
  }

  /**
   * Gets active route
   * @returns active route 
   */
  public getActiveRoute(): IMenuItem {
    return this._menu.getActiveRoute();
  }

  /**
   * Gets previous route
   * @returns previous route 
   */
  public getPreviousRoute(): IMenuItem {
    return this._menu.getPreviousRoute();
  }

  /**
   * Gets breadcrumbs for active route
   * while(parentId !== 0) - I limit this to 3 iterations by default. I can't see why you would need more - but you never know
   * @returns breadcrumbs for active route 
   */
  public getBreadcrumbsForActiveRoute(maxIterations = 3): IMenuItem[] {
    return this._menu.getPathway(this._menu.getActiveRoute(), maxIterations = 3);
  }
  /**
   * Gets menu for role
   * @param parent 
   * @param role 
   */
  public getMenuForRoleAndParent(parent: number, role?: string): IMenuItem[] {

    const children = this._menu.getChildrenForMenuItem(parent);
    return role === undefined ? children :
      children.filter(x => x.roles !== undefined).filter(y => y.roles.some(z => z === role));
  }
  /**
   * Gets menu for role
   * @param parent
   * @param role
   */
  public getTopMenuItems(role?: string): IMenuItem[] {
    const topMenu = this._menu.getAllRoutes().filter(x => x.isTopLevel);
    topMenu.forEach(item => {
      item.children = this.getMenuForRoleAndParent(item.id, role);
    });
    return topMenu;
  }



  /**
   * Gets menu with optional filter by current role
   * Could argue we add authentication provider but I think this should remain implementation agnostic
   * @param [roleFilter] 
   * @returns menu 
   */
  public getMenuRoutes(roleFilter?: string): IMenuItem[] {
    return roleFilter === undefined ?
      this._menu.getAllRoutes() :
      this._menu.getAllRoutes().filter(x => x.roles !== undefined && x.roles.some(y => y === roleFilter));
  }
  public getMenu(): Menu {
    return this._menu;
  }

}
